use log::debug;
use std::sync::Arc;

use crate::overview;

// pub(crate) struct Law {
//     name: String,          //ABGB, UrhG
//     section: Vec<Section>, // § 1, § 2, ...
// }

// impl Law {
//     pub(crate) fn new(name: &str) -> Self {
//         Self {
//             name: name.into(),
//             section: Vec::new(),
//         }
//     }
// }

/// Is used to generate a law struct. It's organized mainly by classifier.
#[derive(Debug)]
pub(crate) struct LawBuilder {
    /// Name of the law
    name: String, //ABGB, UrhG

    /// Structure of the law text
    classifiers: Vec<Classifier>,

    last_header_index: Option<usize>,

    /// Stores the header of the next paragraph
    next_para_header: Option<String>,
}

fn contains(classifier_name: &str, instance_name: &str) -> bool {
    instance_name.contains(classifier_name)
}

fn starts_with_number(_classifier_name: &str, instance_name: &str) -> bool {
    match instance_name.trim().as_bytes().get(0) {
        Some(c) if c.is_ascii_digit() => true,
        _ => false,
    }
}

impl LawBuilder {
    #[cfg(test)]
    pub(crate) fn test(name: &str) -> Self {
        let mut classifiers = Vec::new();
        if name == "UrhG" {
            let hauptstueck = Classifier::new("Hauptstück", Arc::new(&contains));
            classifiers.push(hauptstueck.clone());

            let mut abschnitt = Classifier::new("Abschnitt", Arc::new(&contains));
            abschnitt.set_parent(hauptstueck);
            classifiers.push(abschnitt);

            let mut numbered_header =
                Classifier::new("Numbered Header", Arc::new(&starts_with_number));
            numbered_header.set_parent(abschnitt);
            classifiers.push(numbered_header);
        } else if name == "no-headers" {
            let mut class = Classifier::new("", Arc::new(&contains));
            class.add_instance(ClassifierInstance::new(""));
            classifiers.push(class);
        }
        Self {
            name: name.into(),
            classifiers,
            next_para_header: None,
            last_header_index: None,
        }
    }

    /// Creates a new law builder. Adds classifier for known law texts.
    pub(crate) fn new(name: &str) -> LawBuilder {
        //TODO: return Law (not LawBuilder)
        let mut classifiers = Vec::new();

        let mut law_id = None;
        if name == "UrhG" {
            law_id = Some(10001848);
            let hauptstueck = Classifier::new("Hauptstück", Arc::new(&contains));
            classifiers.push(hauptstueck.clone());

            let mut abschnitt = Classifier::new("Abschnitt", Arc::new(&contains));
            abschnitt.set_parent(hauptstueck);
            classifiers.push(abschnitt.clone());

            let mut numbered_header =
                Classifier::new("Numbered Header", Arc::new(&starts_with_number));
            numbered_header.set_parent(abschnitt);
            classifiers.push(numbered_header);
        }

        let mut builder = Self {
            name: name.into(),
            classifiers,
            next_para_header: None,
            last_header_index: None,
        };

        overview::parse(law_id.unwrap(), &mut builder).unwrap();

        builder
    }

    /// Sets a new header.
    pub(crate) fn new_header(&mut self, name: &str) {
        debug!("new_header={name}");
        let classifier_index = self
            .classifiers
            .iter()
            .position(|class| class.used_for(name));
        match classifier_index {
            Some(index) => {
                let class = ClassifierInstance::new(name);
                self.classifiers[index].add_instance(class);
                self.last_header_index = Some(index);
            }
            None => panic!("No classifier for {name}"),
        }
    }

    /// Sets a new description for the last classifier.
    pub(crate) fn new_desc(&mut self, desc: &str) {
        debug!("new_desc={desc}");
        if let Some(index) = self.last_header_index {
            self.classifiers[index].set_desc(desc);
        } else {
            panic!("Not possible");
        }
    }

    /// Adds a new paragraph.
    pub(crate) fn new_par(&mut self, par: String, content: Content) {
        debug!("new_par=par:{par};content:{content:#?}");
        if let Some(index) = self.last_header_index {
            let section = Section {
                symb: par,
                content,
                par_header: self.next_para_header.clone(),
            };
            self.next_para_header = None;
            self.classifiers[index].add_section(section);
        } else {
            panic!("Expected at least one classifier");
        }
    }

    /// Next paragraph has a header, store its name.
    pub(crate) fn new_next_para_header(&mut self, header: &str) {
        debug!("new_next_para_header={header}");
        self.next_para_header = Some(header.into());
    }

    pub(crate) fn toc(&self) {
        for class in &self.classifiers {
            for inst in &class.instances {
                println!("{}", inst.name);
            }
        }
    }
}

#[derive(Debug, Clone)]
pub(crate) struct Section {
    symb: String, // §"1", §"2", ...
    par_header: Option<String>,
    content: Content,
    //header: Option<Header>,
}

#[derive(Clone)]
pub(crate) struct Header {
    classifier: Classifier, // Hauptstück, Theil, Abschnitt, ol
    name: String,           // 1. Hauptstück, 3. Theil, 7. Abschnitt, li
    parent: Option<Box<Header>>,
}

impl Header {
    fn get_all_classifiers(&self) -> Vec<Classifier> {
        let mut ret = Vec::new();
        ret.push(self.classifier.clone());
        if let Some(parent) = &self.parent {
            for a in parent.get_all_classifiers() {
                ret.push(a);
            }
        }

        ret
    }
}

#[derive(Clone, Debug)]
pub(crate) struct ClassifierInstance {
    name: String,
    desc: Option<String>,
    sections: Vec<Section>,
}

impl ClassifierInstance {
    fn new(name: &str) -> Self {
        Self {
            name: name.into(),
            desc: None,
            sections: Vec::new(),
        }
    }

    fn set_desc(&mut self, desc: &str) {
        self.desc = Some(desc.into());
    }

    fn add_section(&mut self, section: Section) {
        self.sections.push(section);
    }
}

#[derive(Clone)]
pub(crate) struct Classifier {
    name: String, // Hauptstück, Theil, Abschnitt, ol
    parent: Option<Box<Classifier>>,
    instances: Vec<ClassifierInstance>,
    used_for_fn: Arc<dyn Fn(&str, &str) -> bool>,
}

impl std::fmt::Debug for Classifier {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Classifier")
            .field("name", &self.name)
            .field("parent", &self.parent)
            .field("instances", &self.instances)
            .finish()
    }
}

impl Classifier {
    fn new(name: &str, used_for_fn: Arc<dyn Fn(&str, &str) -> bool>) -> Self {
        Self {
            name: name.into(),
            parent: None,
            instances: Vec::new(),
            used_for_fn,
        }
    }

    fn set_parent(&mut self, parent: Classifier) {
        self.parent = Some(Box::new(parent));
    }

    fn contains(&self, name: &str) -> bool {
        name.contains(&self.name)
    }

    fn add_instance(&mut self, name: ClassifierInstance) {
        self.instances.push(name);
    }

    fn set_desc(&mut self, desc: &str) {
        self.instances.last_mut().unwrap().set_desc(desc);
    }

    fn used_for(&self, name: &str) -> bool {
        (self.used_for_fn)(&self.name, name)
    }

    fn add_section(&mut self, section: Section) {
        self.instances.last_mut().unwrap().add_section(section);
    }
}

#[derive(Clone, Debug)]
pub(crate) enum Content {
    Text(String),            //This is my direct law text
    Item(Vec<Box<Content>>), //(1) This is general law. (2) This is more specific law
    List(Vec<Box<Content>>),
}

#[cfg(test)]
mod tests {
    use super::*;
    use pretty_assertions::assert_eq;

    #[test]
    fn test() {
        let mut builder = LawBuilder::test("UrhG");

        builder.new_header("1. Hauptstück");
        builder.new_header("2. Abschnitt");

        builder.new_par(Content::Text("Mein erster Paragraph".into()));

        let expected = LawBuilder {
            name: "UrhG".into(),
            classifiers: vec![
                Classifier {
                    name: "Hauptstück".into(),
                    parent: None,
                    instances: vec![ClassifierInstance {
                        name: "1. Hauptstück".into(),
                        desc: None,
                        content: vec![],
                    }],
                },
                Classifier {
                    name: "Abschnitt".into(),
                    parent: Some(Box::new(Classifier {
                        name: "Hauptstück".into(),
                        parent: None,
                        instances: vec![],
                    })),
                    instances: vec![ClassifierInstance {
                        name: "2. Abschnitt".into(),
                        desc: None,
                        content: vec![Content::Text("Mein erster Paragraph".into())],
                    }],
                },
            ],
            next_para_header: None,
            last_header_index: Some(1),
        };
        assert_eq!(builder, expected);
    }
}
